/*********

08.03.2022: ftp-routine im loop neu plaziert


mit update über ElegantOTA
mit Speicherung im FRAM

Georg Luber
Projektname: E-Datalogger


*********/


#include <Arduino.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <AsyncElegantOTA.h>
#include "SPIFFS.h"
#include "Adafruit_FRAM_SPI.h"
#include <Arduino_JSON.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <Wire.h>
#include "ESP32FtpServer.h"

// Für die WLAN - Verbindung:
const char* ssid     = "Meine_SSID"; // Hier die SSID des WLAN eingeben
const char* password = "WLAN_Passwort";  // und hier das zugehörige Passwort


FtpServer ftpSrv; 

/*
Zur Info: Ansteuerpins der SD-Karte und FRAM:
ESP       //    SD-Karte   
Pin    4  // verbinden mit CS (FRAM)
Pin    5  // verbinden mit CS (SD)
Pin   18  // verbinden mit SCK/CLK
Pin   19  // verbinden mit MISO/SDO
Pin   23  // verbinden mit MOSI/SDI
*/
uint8_t FRAM_CS = 4;

Adafruit_FRAM_SPI fram = Adafruit_FRAM_SPI(FRAM_CS);  // Hardware SPI wird genutzt

// Setze Input für WR Zähleingänge
const int In_WR1 = 33;
const int In_WR2 = 25;
const int In_WR3 = 26;
const int In_WR4 = 27;
const int In_WR5 = 32;
const int LED_Pin=2;

// Globale Variable
unsigned long Istzeit [6];
unsigned long Altzeit [6];
unsigned long HiVar[6];
unsigned int P[6];
unsigned long Impulszaehler[6];
unsigned long Impulsmerker[6];
/* 
Impulszaehler [1] enthält die Impulse von WR 1
Impulszaehler [2] enthält die Impulse von WR 2
usw.
*/
 
unsigned long Zeit1;
unsigned long Zeit2;
unsigned long Vgl=40; //40ms Entprellzeit
unsigned long Diff;

unsigned long t0 = 0;
unsigned long WLAN_testzeit = 30000; // Timeout, falls die WLAN-Verbindung verloren geht

int Eingang[6];

// Für den Zugriff auf die Zeit aus dem Internet
const char* ntpServer = "ptbtime1.ptb.de";
const long  gmtOffset_sec = 3600;
const int   daylightOffset_sec = 3600;


// Für die Aufbereitung von Zeit und Datum

int Jahr;
int Monat;
int Tag;
int Stunde;
int Minute;

bool Neuer_Tag;
bool Starte_Aufzeichnung;
bool Gibts_datei;


int Alt_Jahr=0;
int Alt_Monat=0;
int Alt_Tag=0;
int Alt_Stunde=0;
int Alt_Minute=0;


// Für den Test des Timers

unsigned long Zeit;
unsigned long Altzeit_timer=0;
unsigned long Differenz;

/* legt den hardware timer an*/
hw_timer_t * timer = NULL;

portMUX_TYPE mux = portMUX_INITIALIZER_UNLOCKED; 

// AsyncWebServer auf Port 80
AsyncWebServer server(80);

// Lege WebSocket Objekt an
AsyncWebSocket ws("/ws");

// Lege die Ereignisse an
AsyncEventSource events("/events");
AsyncEventSource event2("/zugriff");

// Json Variable für die Leistung
JSONVar readings;

// Json Variable für die vom ESP gemessene Feldstärke zur Überprüfung der Qualität der WLAN-Verbindung
JSONVar daten;

// Timer Variablen
unsigned long lastTime = 0;
unsigned long timerDelay = 30000;
bool SD_Merker;

// Für die SD
String Filename;
String Verzeichnis;

String Jahr_dir;
String Monat_dir;

File dataFile;
File arbeit;
File root;
const unsigned int STRING_BUFFER_SIZE = 128;
char Zeichenpuffer[STRING_BUFFER_SIZE];


ulong Startzaehler; // zählt die Neustarts des Dataloggers

// definition der Adressen für das FRAM
uint16_t Adr_Startzaehler=0; // in Adresse 0 des FRAM steht die Anzahl der Starts


unsigned long Imp_aktuell; // speichert die aktuellen Gesamtimpulse
unsigned long Imp_tag; // enthält die Impulse des Tages um 00:00 Uhr
unsigned long Imp_vortag; // enhält die Impulse des Vortages um 00:00 Uhr

const uint16_t Adr_Imp_vortag=4; // zugehörige Adressen im FRAM
const uint16_t Adr_Imp_tag=8;
const uint16_t Adr_Imp_aktuell=12;
const uint16_t Adr_Impulszaehler=16;

float Ertrag_vortag;
float Ertrag_tag;

bool Neuer_Wert;
bool Aufruf;

String string_numberfix(int a, int length)
{
String Ergebnis = String(a);
String Ges_String="000000" + Ergebnis; // Falls der String kürzer ist, wird er vorne durch Leerzeichen verlängert.
    int laenge=(Ges_String.length()); // in laenge steht die Länge des Gesamtstrings
    int start=laenge-length; // Der String soll auf "length" Zeichen gekürzt werden
    String ausgabe=Ges_String.substring(start,laenge); // es werden die rechten "length" Zeichen in ausgabe gespeichert
return ausgabe;
}

void FRAM_Schreibe (uint16_t adresse, ulong variable)
{
  // schreibt die ulong Variable in die mitgegebene Adresse
uint8_t fram_byte[4];
uint16_t addr = adresse;

for(int i=0;i<4;i++)
  {
    fram_byte[i]=(variable>>(i*8)) & 0xFF;
    // Schreibe ins FRAM
    fram.writeEnable(true);
    fram.write8(addr, fram_byte[i]);
    fram.writeEnable(false);
    addr++;
  }
}

ulong FRAM_Lese (uint16_t adresse)
{
// liest ulong Variable ab der mitgegebnenen Adresse
//   uint8_t test = fram.read8(0x0);

  uint16_t addr=adresse;
  uint8_t  gelesen[4];
  ulong erg=0; 

for(int i=0;i<4;i++)
{
  gelesen[i]=fram.read8(addr);
  erg=erg+(gelesen[i]<<(i*8));
  addr++;
}
  return (erg); 
}


// Hier folgt der Timerinterrupt:
void IRAM_ATTR onTimer()
{

  // Ausgabe für Testzwecke
   Zeit=millis();
   Differenz=Zeit-Altzeit_timer;
   // Serial.println(Differenz);
   
   Altzeit_timer=Zeit;
   SD_Merker=true;
   for (int n=1; n<6; n++)
    {
      Impulsmerker[n]=Impulszaehler[n]; // merke die Impulse für die spätere Aufzeichnung auf SD. Später, da die Interruptroutine möglichst kurz sein soll
    }

}

// Interruptroutine für WR1

void IRAM_ATTR Imp_WR1 () 
{
  portENTER_CRITICAL(&mux);
  
  // Entprellroutine für Input
 Zeit1=millis();
 // warte die Vgl-Zeit ab:
 do {
   Zeit2=millis();
   Diff=Zeit2-Zeit1;
} while (Diff<Vgl);

 Eingang[1] = digitalRead(In_WR1);  // Liest den Inputpin -> Lowaktiv!
 // falls er nach Vgl-Zeit in ms noch Low ist, dann zählt der Impuls.
 
 if (Eingang[1]==0)
 {
  Neuer_Wert=true;
  
  Impulszaehler[1]++; // Zaehlt die Impulse von WR1

  // zum testen:
  // Serial.print("Impulszaehler[1] = ");
  // Serial.println(Impulszaehler[1]);

  Altzeit[1]=Istzeit[1];
  // Istzeit[1] = millis();
  Istzeit[1] = Zeit1;
  HiVar[1] = Istzeit[1]-Altzeit[1];

  if (HiVar[1]>0)
  {
  P[1]=int (1.8E6/HiVar[1]); // Berechnung der Leistung aus dem Zeitabstand zwischen 2 Impulsen.
  }
 }
      portEXIT_CRITICAL(&mux);   
}



// Interruptroutine für WR2

void IRAM_ATTR Imp_WR2 () 
{
  portENTER_CRITICAL(&mux);
  
  // Entprellroutine für Input
 Zeit1=millis();
 // warte die Vgl-Zeit ab:
 do {
   Zeit2=millis();
   Diff=Zeit2-Zeit1;
} while (Diff<Vgl);

 Eingang[2] = digitalRead(In_WR2);  // Liest den Inputpin -> Lowaktiv!
 // falls er nach Vgl-Zeit in ms noch Low ist, dann zählt er:
 
 if (Eingang[2]==0)
 {    
  Neuer_Wert=true;

  Impulszaehler[2]++; // Zaehlt die Impulse von WR2
  // zum testen:
  // Serial.print("Impulszaehler[2] = ");
  // Serial.println(Impulszaehler[2]);


  Altzeit[2]=Istzeit[2];
  //Istzeit[2] = millis();
  Istzeit[2] = Zeit1;
  HiVar[2] = Istzeit[2]-Altzeit[2];

  if (HiVar[2]>0)
  {
  P[2]=int (1.8E6/HiVar[2]); // Berechnung der Leistung aus dem Zeitabstand zwischen 2 Impulsen.
  }

 }
      portEXIT_CRITICAL(&mux); 
}



// Interruptroutine für WR3

void IRAM_ATTR Imp_WR3 () 
{

  portENTER_CRITICAL(&mux);
  
  // Entprellroutine für Input
 Zeit1=millis();
 // warte die Vgl-Zeit ab:
 do {
   Zeit2=millis();
   Diff=Zeit2-Zeit1;
} while (Diff<Vgl);

 Eingang[3] = digitalRead(In_WR3);  // Liest den Inputpin -> Lowaktiv!
 // falls er nach Vgl-Zeit in ms noch Low ist, dann zählt er:
 
 if (Eingang[3]==0)
 {    
  Neuer_Wert=true;

  Impulszaehler[3]++; // Zaehlt die Impulse von WR3

  // zum testen:
  // Serial.print("Impulszaehler[3] = ");
  // Serial.println(Impulszaehler[3]);


  Altzeit[3]=Istzeit[3];
  //Istzeit[3] = millis();
  Istzeit[3] = Zeit1;
  HiVar[3] = Istzeit[3]-Altzeit[3];

  if (HiVar[3]>0)
  {
  P[3]=int (1.8E6/HiVar[3]); // Berechnung der Leistung aus dem Zeitabstand zwischen 2 Impulsen.
  }
 }
      portEXIT_CRITICAL(&mux); 
}



// Interruptroutine für WR4

void IRAM_ATTR Imp_WR4 () 
{
  portENTER_CRITICAL(&mux);
    
  // Entprellroutine für Input
 Zeit1=millis();
 // warte die Vgl-Zeit ab:
 do {
   Zeit2=millis();
   Diff=Zeit2-Zeit1;
} while (Diff<Vgl);

 Eingang[4] = digitalRead(In_WR4);  // Liest den Inputpin -> Lowaktiv!
 // falls er nach Vgl-Zeit in ms noch Low ist, dann zählt er:
 
 if (Eingang[4]==0)
 {     
  Neuer_Wert=true;

  Impulszaehler[4]++; // Zaehlt die Impulse von WR4

  // zum testen:
  // Serial.print("Impulszaehler[4] = ");
  // Serial.println(Impulszaehler[4]);


  Altzeit[4]=Istzeit[4];
  //Istzeit[4] = millis();
  Istzeit[4] = Zeit1;
  HiVar[4] = Istzeit[4]-Altzeit[4];

  if (HiVar[4]>0)
  {
  P[4]=int (1.8E6/HiVar[4]); // Berechnung der Leistung aus dem Zeitabstand zwischen 2 Impulsen.
  } 
 }
      portEXIT_CRITICAL(&mux); 
     
      
}

// Interruptroutine für WR5

void IRAM_ATTR Imp_WR5 () 
{
  portENTER_CRITICAL(&mux);
    
  // Entprellroutine für Input
 Zeit1=millis();
 // warte die Vgl-Zeit ab:
 do {
   Zeit2=millis();
   Diff=Zeit2-Zeit1;
} while (Diff<Vgl);

 Eingang[5] = digitalRead(In_WR5);  // Liest den Inputpin -> Lowaktiv!
 // falls er nach Vgl-Zeit in ms noch Low ist, dann zählt er:
 
 if (Eingang[5]==0)
 {     
  Neuer_Wert=true;

  Impulszaehler[5]++; // Zaehlt die Impulse von WR4

  // zum testen:
  // Serial.print("Impulszaehler[5] = ");
  // Serial.println(Impulszaehler[5]);


  Altzeit[5]=Istzeit[5];
  Istzeit[5] = Zeit1;
  HiVar[5] = Istzeit[5]-Altzeit[5];

  if (HiVar[5]>0)
  {
  P[5]=int (1.8E6/HiVar[5]); // Berechnung der Leistung aus dem Zeitabstand zwischen 2 Impulsen.
  } 
 }
      portEXIT_CRITICAL(&mux); 
     
      
}

void Aufzeichnung_SD()
{
  if (Starte_Aufzeichnung==true)
  {
    if (SD_Merker==true)
    {
    // hier werden die Daten, welche beim Timerinterrupt gemerkt wurden, auf die SD geschrieben
      
      char fname[20];

      Filename.toCharArray(fname, 16); 
      dataFile = SD.open(fname, FILE_APPEND); // öffne Datei zum Schreiben
      if (Aufruf==false)
      {
        // es wird das erste mal geschrieben
        String stunde_str=string_numberfix(Stunde,2);
        dataFile.print(stunde_str);
        dataFile.print(":");
        String minute_str=string_numberfix(Minute,2);
        dataFile.println(minute_str);
        dataFile.println(">");
        Aufruf=true; // Zeit wird nur 1x eingetragen
        for (int n=1; n<5; n++)
        {
          dataFile.print (Impulsmerker[n]);
          dataFile.print (",");
        }
      dataFile.println (Impulsmerker[5]);
      }
      else
      {
      for (int n=1; n<5; n++)
        {
          dataFile.print (Impulsmerker[n]);
          dataFile.print (",");
        }
      dataFile.println (Impulsmerker[5]);
      }
            
      dataFile.close();

      SD_Merker=false; // es wurde aufgezeichnet, der Timer gibt vor, wann wieder aufgezeichnet wird

    }
  }
}


String convertToString(char* a, int size)
// wandelt char-array in String
{
    int i;
    String s = "";
    for (i = 0; i < size; i++) {
        s = s + a[i];
    }
    return s;
}

// Hole Daten und schreibe sie ins JSON Objekt
String holWR_Daten(){
  readings["p_ges"] = String(P[0]);
  readings["wr_1"] = String(P[1]);
  readings["wr_2"] = String(P[2]);
  readings["wr_3"] = String(P[3]);
  readings["wr_4"] = String(P[4]); 
  readings["wr_5"] = String(P[5]); 
  readings["ertr_tg"] = String(Ertrag_tag,2);
  readings["ertr_vg"] = String(Ertrag_vortag,2);
  String jsonString = JSON.stringify(readings);
  return jsonString;
}

// Lese die WLAN Feldstärke und aktualisiere das JSON Objekt
String HolWlanFS(){
  daten["feldst"] =  String(WiFi.RSSI());
  daten["startz"] =  String(Startzaehler);
  String ergebnis = JSON.stringify(daten);
  return ergebnis;
}


// Initialisiere SPIFFS
void initSPIFFS() {
  if (!SPIFFS.begin()) {
    Serial.println("Fehler beim Aufbau des SPIFFS");
  }
  Serial.println("SPIFFS erfolgreich aufgebaut");
}

// Initialize WiFi
void initWiFi() {
  // WiFi.mode(WIFI_STA);
/*
  IPAddress ip(192, 168, 178, 240);
  IPAddress dns(192, 168, 178, 1);
  IPAddress gateway(192, 168, 178, 1);
  IPAddress subnet(255, 255, 255, 0);
  WiFi.config(ip, dns, gateway, subnet);
  */

  WiFi.begin(ssid, password);
  Serial.print("Verbinde mit WLAN ..");
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    delay(500);
  }
  Serial.println(WiFi.localIP());

}


void notifyClients(String state) {
  ws.textAll(state);
}



void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client,AwsEventType type,
             void *arg, uint8_t *data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT:
      Serial.printf("WebSocket client #%u connected from %s\n", client->id(), client->remoteIP().toString().c_str());
      break;
    case WS_EVT_DISCONNECT:
      Serial.printf("WebSocket client #%u disconnected\n", client->id());
      break;
    
    case WS_EVT_DATA:
      break;
    
    case WS_EVT_PONG:
    case WS_EVT_ERROR:
      break;
  }
}

void initWebSocket() {
    ws.onEvent(onEvent);
    server.addHandler(&ws);
}




bool zeile_lesen(File& file)
{
  static unsigned int index;

  while (file.available())
  {
    char c = file.read();

    if (c > 32 && index < STRING_BUFFER_SIZE - 1)
    {
      Zeichenpuffer[index++] = c;
    }
    else if (c == '\n')
    {
      Zeichenpuffer[index] = '\0';
      index = 0;
      return true;
    }
  }

  return false;
}


void printLocalTime() // wird nur für's Debugging benötigt
{
  struct tm timeinfo;
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time");
    return;
  }
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S");
  Serial.print("Day of week: ");
  Serial.println(&timeinfo, "%A");
  Serial.print("Month: ");
  Serial.println(&timeinfo, "%B");
  Serial.print("Day of Month: ");
  Serial.println(&timeinfo, "%d");
  Serial.print("Year: ");
  Serial.println(&timeinfo, "%Y");
  Serial.print("Hour: ");
  Serial.println(&timeinfo, "%H");
  Serial.print("Hour (12 hour format): ");
  Serial.println(&timeinfo, "%I");
  Serial.print("Minute: ");
  Serial.println(&timeinfo, "%M");
  Serial.print("Second: ");
  Serial.println(&timeinfo, "%S");

  Serial.println("Time variables");
  char timeHour[3];
  strftime(timeHour,3, "%H", &timeinfo);
  Serial.println(timeHour);
  char timeWeekDay[10];
  strftime(timeWeekDay,10, "%A", &timeinfo);
  Serial.println(timeWeekDay);
  Serial.println();
}




void Hol_Zeit_Variablen(){
  struct tm timeinfo;
    if(!getLocalTime(&timeinfo)){
// Zeitfehler
    return;
  }
  char h_stunde[3];
  strftime(h_stunde,3, "%H", &timeinfo);

  String stunde = convertToString(h_stunde, 3);
  Stunde=stunde.toInt();


  char h_minute[3];
  strftime(h_minute,3, "%M", &timeinfo);
  String minute = convertToString(h_minute, 3);
  Minute=minute.toInt();


  char h_jahr[5];
  strftime(h_jahr,5, "%Y", &timeinfo);
  String jahr = convertToString(h_jahr, 4);
  Jahr_dir="/"+jahr; // legt das Jahresverzeichnis fest
  Jahr=jahr.toInt();
 
  
  char h_monat[3];
  strftime(h_monat,3, "%m", &timeinfo);
  String monat = convertToString(h_monat, 2);
  Monat_dir="/"+monat; // legt das Monatsverzeichnis fest
  Monat=monat.toInt();

  
  char h_tag[3];
  strftime(h_tag,3, "%d", &timeinfo);
  String tag = convertToString(h_tag, 2);
  Tag=tag.toInt();


  // Es wird das Verzeichnis und der Filneame für das Datenfile gebildet:
  Verzeichnis="/"+jahr + "/" + monat;
  Filename=Verzeichnis+"/"+tag+".csv";
}



void setup() 
{
  // Serial Port zum debuggen
  Serial.begin(115200);
  pinMode(LED_Pin, OUTPUT);
  digitalWrite(LED_Pin, LOW);
  initWiFi();
  Serial.print("WLAN-Feldstärke: ");
  Serial.println(WiFi.RSSI());

  // Init and get the time
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
  // printLocalTime();
  // Hol_Zeit_Variablen();

  
    // Eingänge für Impulszählung der WR 1 bis 4
  pinMode(In_WR1, INPUT_PULLUP);
  pinMode(In_WR2, INPUT_PULLUP);
  pinMode(In_WR3, INPUT_PULLUP);
  pinMode(In_WR4, INPUT_PULLUP);
  pinMode(In_WR5, INPUT_PULLUP);

  // Interrupts für die WR 1 bis 5 einrichten mit Triggerung auf fallende Flanke

  attachInterrupt(digitalPinToInterrupt(In_WR1), Imp_WR1, FALLING);

  attachInterrupt(digitalPinToInterrupt(In_WR2), Imp_WR2, FALLING);

  attachInterrupt(digitalPinToInterrupt(In_WR3), Imp_WR3, FALLING);

  attachInterrupt(digitalPinToInterrupt(In_WR4), Imp_WR4, FALLING);

  attachInterrupt(digitalPinToInterrupt(In_WR5), Imp_WR5, FALLING);

  /* Timerinterrupt auf 300s=5min stellen
  Nimm 1. timer (von 4)
  1 tick ist 1/(80MHZ/8000) = 100us - Teiler auf 8000 und hochzählen */

  timer = timerBegin(0, 8000, true);

  /* Timer und Interrupt einrichten */
  timerAttachInterrupt(timer, &onTimer, true);

  // Alarm auslösen, fall der Timer 3000000 erreicht hat das sind 3E6

  timerAlarmWrite(timer, 3E6, true);

  /* Starte den Timerinterrupt */
  timerAlarmEnable(timer);



  initSPIFFS();
  // Serial.println("SPIFFS gestartet");


  // Hole die Zeit aus dem Internet
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
  
  Hol_Zeit_Variablen();


  
    if(!SD.begin())
    {
        Serial.println("keine SD-Karte, bzw. SD-Problem");
        //Fehlerbearbeitung und ggf. Meldung
        while(1) // endlose while Schleife; 1 ist immer wahr
          {
            // Prozessor bleibt in der Endlosschleife, LED an Pin 2 blinkt
            digitalWrite(LED_Pin, HIGH); // lässt die LED doppelblinken: SD-Fehler  
            delay (900);
            digitalWrite(LED_Pin, LOW);
            delay (100);
            digitalWrite(LED_Pin, HIGH);  
            delay (100);
            digitalWrite(LED_Pin, LOW);
            delay (100);
            // Abhilfe: SD-Karte einlegen bzw. Fehler beseitigen - anschließend manuell Reset auslösen 
          }
    }             

    uint8_t cardType = SD.cardType();

    if(cardType == CARD_NONE)
      {
        Serial.println("Keine SD - Karte");
        return;
      }

        Serial.print("SD Card Type: ");
    if(cardType == CARD_MMC){
        Serial.println("MMC");
    } else if(cardType == CARD_SD){
        Serial.println("SDSC");
    } else if(cardType == CARD_SDHC){
        Serial.println("SDHC");
    } else {
        Serial.println("UNKNOWN");
    }

    uint64_t cardSize = SD.cardSize() / (1024 * 1024);
    Serial.printf("SD Card Size: %lluMB\n", cardSize);

    ftpSrv.begin("esp32","esp32");    //Benutzername und Passwort für FTP.
  
  if (fram.begin()) {
    Serial.println("SPI FRAM gefunden");
  } else {
    Serial.println("Kein SPI FRAM entdeckt ... bitte die Verbindung überprüfen\r\n");
    while (1);
  }
 
 /*
  ulong Startzaehler = FRAM_Lese (Adr_Startzaehler); // lese den Startzähler
  Serial.print("Startzähler = ");
  Serial.println(Startzaehler);
  Serial.println(" ");
  Startzaehler++;
  FRAM_Schreibe (Adr_Startzaehler, Startzaehler); // erhöhe den Startzähler um 1
*/

  ulong startz = FRAM_Lese (Adr_Startzaehler); // lese den Startzähler
  Serial.print("Startzähler = ");
  Serial.println(startz);
  Serial.println(" ");
  startz++;
  FRAM_Schreibe (Adr_Startzaehler, startz); // erhöhe den Startzähler um 1
  Startzaehler=startz;

  // Hole die Daten für den aktuellen und den gestrigen Verbrauch aus dem FRAM.
  // lese die Daten der aktuellen Impulse und die Impulse um Mitternacht und um Mitternacht des Vortages//
  Imp_aktuell=FRAM_Lese(Adr_Imp_aktuell);
  Imp_tag=FRAM_Lese(Adr_Imp_tag);
  Imp_vortag=FRAM_Lese(Adr_Imp_vortag);

  Serial.println("Setup: Impulse vom FRAM:");
  Serial.print("Imp_aktuell = ");
  Serial.println(Imp_aktuell);

  Serial.print("Imp_tag = ");
  Serial.println(Imp_tag);

  Serial.print("Imp_vortag = ");
  Serial.println(Imp_vortag);


  // Berechne die Daten
  Ertrag_tag=float (Imp_aktuell-Imp_tag)*0.0005; // in kWh bei 0,5 Wh pro S0-Impuls
  Ertrag_vortag=float (Imp_tag-Imp_vortag)*0.0005; // ebenfalls in kWh
  Serial.print("Ertrag_tag = ");
  Serial.println(Ertrag_tag);  
  Serial.print("Ertrag_vortag = ");
  Serial.println(Ertrag_vortag);

  uint16_t adr_imp=Adr_Impulszaehler;
  for (int i = 1; i < 6; i++)
  {
    Impulszaehler[i]=FRAM_Lese (adr_imp);
    Serial.print("adr_imp = ");
    Serial.print(adr_imp);
    Serial.print(",    ");
    Serial.print("Impulszaehler [");
    Serial.print(i);
    Serial.print("] = ");
    Serial.println(Impulszaehler[i]);

    adr_imp=adr_imp+4; // Es sind immer 4 Bytes pro ulong Eintrag erforderlich
  }


  SD_Merker=false; // beim Start gibt es keine Aufzeichnung. Freischaltung erfolgt in der Timerroutine

  for (int i = 0; i < 6; i++)
  {
    P[i]=0; // Alle Leistungen der WR zum Start auf 0 setzen
  }

  initWebSocket();

  server.serveStatic("/", SPIFFS, "/");

  // Web Server Root URL
  // Route zur /outputs Webseite
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/messwerte.html", "text/html",false);
  });


  server.on("/about", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/about.html", "text/html");
  });


  // Anfrage nach WR-Leistung
  server.on("/readings", HTTP_GET, [](AsyncWebServerRequest *request){
    String json = holWR_Daten();
    request->send(200, "application/json", json);
    json = String();
  });
  
  // Anfrage nach Feldstärke für die Info-Seite
  server.on("/holdaten", HTTP_GET, [](AsyncWebServerRequest *request) {
    String json = HolWlanFS();
    request->send(200, "application/json", json);
    json = String();
  });
 


  events.onConnect([](AsyncEventSourceClient *client){
    if(client->lastId()){
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    client->send("hello!", NULL, millis(), 10000);
  });
  server.addHandler(&events);
  server.addHandler(&event2);
  AsyncElegantOTA.begin(&server);    // Starte ElegantOT für's Update über WLAN

  // Start server
  server.begin();
  digitalWrite(LED_Pin, HIGH);

  Serial.println("*** Setup-Ende ***");
}

 
void loop() 
{

char fname[20];
AsyncElegantOTA.loop(); // Für OTA Updates

  // Leistung auf Null setzen, falls für 30min.(=1.8€6ms) kein Impuls mehr kommt
// Das sorgt dafür, dass bei Dunkelheit für P=0 bei den WR 1-5 angezeigt wird
  for (int n=0;n<6;n++)
  {
    unsigned long aktzeit=millis(); 
    
    if ((aktzeit-Istzeit[n]) > 1.8E6) // in Istzeit [n] steht die Zeit des letzten empfangenen Impulses des Zählers 
    { 
      P[n]=0;
      P[0]=P[1]+P[2]+P[3]+P[4]+P[5];
      Istzeit[n]=aktzeit; // verhindert, dass ab 30 min dauernd die Anzeige (fals vorhanden) aufgerufen wird
    }
  
  }

    P[0]=0; // Gesamtleistung löschen
    // Berechnung der Gesamtleistung
    P[0]=P[1]+P[2]+P[3]+P[4]+P[5];





  if (Neuer_Wert==true)
    {
    Starte_Aufzeichnung=true; //starte Aufzeichnung
     // Berechnung der aktuellen Gesamtimpulse falls es eine Änderung ergab

    Imp_aktuell=Impulszaehler[1]+Impulszaehler[2]+Impulszaehler[3]+Impulszaehler[4]+Impulszaehler[5];
    FRAM_Schreibe(Adr_Imp_aktuell,Imp_aktuell); // Sichere den Wert im FRAM

    uint16_t adr_imp=Adr_Impulszaehler;

    for (int i = 1; i < 6; i++)
      {
        FRAM_Schreibe (adr_imp,Impulszaehler[i]);
          Serial.print("adr_imp = ");
          Serial.print(adr_imp);
          Serial.print(",    ");
          Serial.print("Impulszaehler [");
          Serial.print(i);
          Serial.print("] = ");
          Serial.println(Impulszaehler[i]);
        adr_imp=adr_imp+4; // 4 Bytes pro ulong Eintrag

      }
    

    Ertrag_tag=float (Imp_aktuell-Imp_tag)*0.0005; // Ertrag des aktuellen Tages
    Neuer_Wert=false;  // Die Daten sind im FRAM, warten auf neue Daten

      // falls neuer Impuls empfangen: gib die aktuellen Werte aus:
      Serial.print("Ertrag = ");
      Serial.println(Ertrag_tag);
      Serial.print("Gesamtimpulse = ");
      Serial.println(Imp_aktuell);

     }
  
  Hol_Zeit_Variablen();

  if (Jahr!=Alt_Jahr)
  {
    // es ist ein neues Jahr erkannt
    // erstelle ein neues Verzeichnis
    // Serial.println ("Jahr!=Alt_Jahr aufgerufen.");
    if (!SD.exists(Jahr_dir)) 
      {
      // Serial.println("Directory Jahr gibt's nicht.");
      // Serial.println("Wird jetzt angelegt.");
      SD.mkdir(Jahr_dir);
      }
   else
      {
      // Serial.println("Directory Jahr ist bereits angelegt.");  
      }
  Alt_Jahr=Jahr;
  }

  if (Monat!=Alt_Monat)
  {
    // es ist ein neues Monat erkannt
    // erstelle in neues Verzeichnis unter Jahr
    // Serial.println ("Monat!=Alt_Monat aufgerufen.");
  char vname[20];
  Verzeichnis.toCharArray(vname, 9);
  if (!SD.exists(vname)) 
      {
      // Serial.println("Directory Monat gibt's nicht.");
      // Serial.println("Wird jetzt angelegt.");
      SD.mkdir(vname);
      }
   else
      {
      // Serial.println("Directory Monat ist bereits angelegt.");  
      }
    Alt_Monat=Monat;
  }


if (Tag!=Alt_Tag)
  {
    // es ist ein neuer Tag erkannt
    // Serial.println("Tag!=Alt_Tag ist aufgerufen");
    Neuer_Tag=true;
    Aufruf=false; // Eintrag Uhrzeit in das SD-File wird vorbereitet für den ersten Dateneintrag
    Alt_Tag=Tag;
    Neuer_Wert=false;
    Starte_Aufzeichnung=false;  // Aufzeichnung startet nach Mitternachterst erst wieder falls ein neuer S0-Impuls empfangen wird. Bis dahin wird Platz auf der SD gespart.

    // Neue Datei auf SD anlegen, Verzeichnis und Filname wurden in "Hol_Zeit_Variablen()" bereits ermittelt. Beides sind Strings

    Filename.toCharArray(fname, 16); 

    dataFile = SD.open(fname, FILE_READ); // Öffne Datei zum lesen, nur zum testen ob die Datei bereits existiert. FILE_READ legt keine Datei an!
            if(!dataFile)
            {
              Gibts_datei=false;
              // Serial.println(fname);
              // Serial.println(" gibt's nicht");  
            }
            else{
              Gibts_datei=true;
              // Serial.println(fname);
              // Serial.println(" existiert");
              dataFile.close();              

            }
 
   
    if (Gibts_datei==false)
    {
      // File gibt's noch nicht, dann lege es an und öffne zum schreiben:

      dataFile = SD.open(fname, FILE_WRITE);
      dataFile.print(Filename); // Schreibe das Datum in die Datei
      dataFile.print(" - ");
      String stunde_str=string_numberfix(Stunde,2);
      dataFile.print(stunde_str);
      dataFile.print(":");
      String minute_str=string_numberfix(Minute,2);
      dataFile.println(minute_str);
      //schreibe den ersten Eintrag mit den letzten Zahlen des Vortages
          for (int m=1; m<6; m++)
            {
            dataFile.print(Impulszaehler[m]);
            dataFile.print (",");    
            }

        dataFile.println(" ");
        dataFile.close();  

        // Bereite die Variablen für die Erfassung der Maxwerte vor
        Imp_vortag=Imp_tag; // es wird um Mitternacht der Vortag angelegt
        Imp_tag=Imp_aktuell; // es werden die aktuellen Daten des Tages um Mitternacht für den Tag gemerkt.
        Ertrag_vortag=(Imp_tag-Imp_vortag)*0.0005; // der aktuelle Tagesertrag wird um Mitternacht zum Vortag gemacht
        Ertrag_tag=0; // der Tagesertrag wird um Mitternacht auf 0 gesetzt

        // Sichere die Daten ins FRAM
        FRAM_Schreibe(Adr_Imp_vortag,Imp_vortag);
        FRAM_Schreibe(Adr_Imp_aktuell,Imp_aktuell);
        FRAM_Schreibe(Adr_Imp_tag,Imp_tag);

    }
  }

  if (Stunde!=Alt_Stunde)
  {
    // Synchronisiere jede Stunde die Zeit mit NTP -> sorgt für gute Genauigkeit
    Alt_Stunde=Stunde;
    configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);
    
  }    

  if ((millis() - lastTime) > timerDelay) 
  {
    events.send("ping",NULL,millis());
    events.send(holWR_Daten().c_str(),"new_readings" ,millis());
    event2.send(HolWlanFS().c_str(),"neuer_zugriff" ,millis());     
    lastTime = millis();
  }
 
    Aufzeichnung_SD();
    events.send("ping",NULL,millis());
    events.send(holWR_Daten().c_str(),"new_readings" ,millis());
    event2.send(HolWlanFS().c_str(),"neuer_zugriff" ,millis()); 

    ftpSrv.handleFTP();


  // falls die Verbindung zum WLAN verloren ging, wird hiermit erneut verbunden
  unsigned long t1 = millis();
  if ((WiFi.status() != WL_CONNECTED) && (t1 - t0 >=WLAN_testzeit)) {
    Serial.println("Erneute Verbindung zum WLAN...");
    WiFi.disconnect();
    WiFi.reconnect();
    t0 = t1;
  }

digitalWrite(LED_Pin, HIGH); // lässt die LED aufblitzen als Zeichen dass die Software läuft  
delay (250);
digitalWrite(LED_Pin, LOW);
}
